package org.steveshipway.blueprints;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.io.Writer;
import java.util.Collections;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.apache.logging.log4j.Logger;
import org.steveshipway.blueprints.directive.*;
import org.steveshipway.blueprints.lib.Constants;

import net.minecraft.block.Block;
import net.minecraft.init.Blocks;
import net.minecraft.world.World;
import net.minecraft.world.chunk.Chunk;

// Minecraft 1.7
import org.steveshipway.blueprints.util17.*;
// Minecraft 1.8
//import org.steveshipway.blueprints.util18.*;
//import net.minecraft.block.state.IBlockState;
//import net.minecraft.util.BlockPos;

public class Blueprint {
	private boolean floaty = false;  // float y coordinate before build
	private boolean dry = false;     // float on water
	private boolean recurse = false; // allow recursion
	private int flatX = 0, flatZ = 0; // flatten this area (float implied)
	private int flatWidth = 0, flatLength = 0; // flatten this area (float implied)
	private int flatMax = 255;         // maximum amount to flatten by
	public boolean visible = true;   // hidden or visible blueprint
	private LinkedList steps = null; // build steps
	public String name;              // my name
	public String helptext = "No description available";
	public float auto_chance = 0;
	public Set biomes = new HashSet(); 
	private Logger logger = null;
	private HashSet<Block> softlist = new HashSet<Block>();
	
	public Blueprint(String name, String description, Boolean v, LinkedList steps, Logger logger) {
		Pattern p;
		Matcher m;
		// create this blueprint
		this.name = name;
		this.steps = steps;
		this.visible = v;
		this.logger = logger;
		if( description != null && description != "" ) {
			this.helptext = description;
		}
		// set up the default softblock list 
		this.softlist.add(Blocks.air);
		this.softlist.add(Blocks.water);
		this.softlist.add(Blocks.flowing_water);
		this.softlist.add(Blocks.lava);
		this.softlist.add(Blocks.flowing_lava);
		
		if( steps.isEmpty() ) {
        	logger.warn("["+name+"] Created a blueprint with no steps?");
        	return;
		}
		
		p = Pattern.compile("(?i)flatten\\s+(-?\\d+)[\\s,]+(-?\\d+)[\\s,]+(-?\\d+)[\\s,]+(-?\\d+)([\\s,]+(\\d+))?.*");
		// iterate over the commands, removing any options and setting the flags
		Iterator i = this.steps.iterator();
	    while(i.hasNext()) {
	        Directive d = (Directive)i.next();
	        String s = d.description();
	        if( s != null ) {
	        	this.helptext = s;
	        	// this.steps.remove(d);
	        	continue;
	        }
	        s = d.option();
	        if( s != null ) {
	        	if( s.matches("(?i)float.*")) { this.floaty = true; continue; }
	        	if( s.matches("(?i)(sink|fixed|static).*")) { this.floaty = false; continue; }
	        	if( s.matches("(?i)dry.*")) {
	        		this.dry = true;
	        		this.floaty = true;
	        		continue;
	        	}
	        	if( s.matches("(?i)recurs.*")) { // catch recurse and recursive
	        		this.recurse = true;
	        		continue;
	        	}
	        	if( s.matches("(?i)(add)?soft .*") ) {
	        		if( ! s.matches("(?i)addsoft.*") ) {
	        			this.softlist.clear();
	        		}
	        		for(String blockname : s.split("[\\s,]+") ) {
	        			if( blockname.matches("(?i)soft") ) { continue; }
	        			if( ! blockname.contains(":") ) { blockname = "minecraft:"+blockname; }
	        			Block b = Block.getBlockFromName(blockname);
	        			if( b == null ) {
	        				logger.warn("Invalid block name in OPTION SOFT directive: "+blockname);
	        			} else {
	        				this.softlist.add(b);
	        			}
	        		}
	        		continue;
	        	}
	        	// option flatten x1,z1  x2,z2 [max]
	        	m = p.matcher(s);
	        	if( m.find() ) {
	        		this.floaty = true;
	        		try {
	        			this.flatX = Integer.parseInt( m.group(1) ); 
	        			this.flatZ = Integer.parseInt( m.group(2) );
	        			this.flatWidth  = Integer.parseInt( m.group(3) ) - this.flatX; 
	        			this.flatLength = Integer.parseInt( m.group(4) ) - this.flatZ;
	        			if( m.group(6) != null ) {
		        			this.flatMax = Integer.parseInt( m.group(6) ); 	        				
	        			}
	        		} catch( Exception e ) {
        				logger.error("Unable to parse OPTION FLATTEN arguments");	 
        				continue;
	        		}
	        		if( this.flatWidth < 0 ) {
	        			this.flatX = this.flatX + this.flatWidth;
	        			this.flatWidth = -this.flatWidth;
	        		}
	        		if( this.flatLength < 0 ) {
	        			this.flatZ = this.flatZ + this.flatLength;
	        			this.flatLength = -this.flatLength;
	        		}
        			//logger.info("Setting flatten area: " + this.flatWidth + "x" + this.flatLength );

	        		continue;
	        	}
	        	// this.steps.remove(d);
				logger.warn("Unable to parse option: " + s);	 	        	
	        	continue;
	        }
	        float f = d.auto_chance();
	        if( f > 0 && f <= 1 ) {
	        	this.auto_chance = f;
	        	this.biomes = ((AutoDirective)d).biomes;
	        	if( ! Portfolio.quiet ) {
	        		logger.info("Added an auto-generated plan '"+this.name+"', chance "+Math.floor(this.auto_chance*Constants.CHANCE_BASE)+"/"+Constants.CHANCE_BASE);
	        	}
	        	continue;
	        } else if( f > 1 ) {
	        	logger.warn("["+this.name+"] Invalid auto chance "+this.auto_chance);
	        }
	    }
    	//logger.info("Created a blueprint with "+this.steps.size()+" steps");
	}
	
	public boolean isVisible() {
		return visible;
	}
	public boolean setVisible(boolean v) {
		visible = v;
		return visible;
	}

	public String helpText() {
		if( this.helptext != null ) {
			return this.helptext;
		}
		return "No description available";
	}
	public String name() { return this.name; }
	
	// build this blueprint, at the specified coordinates.
	public boolean build(World world, BlockPos location, int r ) {
		// iterate over steps, actioning each in turn
		// may require (gulp) recursion...
		// return true if it was successfully built
		Set<Block> floatthrough = new HashSet<Block>();
		boolean rv = true;
		
		if( world.isRemote ) {
        	logger.warn("Blueprint build called with remote world!  Should never happen.");
			return false;			
		}
		if( this.steps == null || this.steps.size() == 0 ) {
        	logger.warn("Build of '"+this.name+"' had no steps!  This seems wrong...");
			return false;
		}
    	logger.debug("Build of blueprint '"+this.name+"' starting: "+this.steps.size()+" steps");

    	if( this.floaty || this.flatWidth > 0 ) {
			if( Portfolio.config.get("blueprint", "floatThroughAir", "true").getBoolean(true) ) {
	       		//if( ! Portfolio.quiet ) { logger.info("Adding Air block to floatable block list"); }
				floatthrough.add(Blocks.air); // treat air as air
			}
			if( Portfolio.config.get("blueprint", "floatThroughPlants", "true").getBoolean(true) ) {
	       		//if( ! Portfolio.quiet ) { logger.info("Adding plants to floatable block list"); }
	    		floatthrough.add(Blocks.log); // treat trees as air
	    		floatthrough.add(Blocks.log2);
	    		floatthrough.add(Blocks.leaves);
	    		floatthrough.add(Blocks.leaves2);
	    		floatthrough.add(Blocks.yellow_flower); // treat flowers and plants as air
	    		floatthrough.add(Blocks.red_flower);
	    		floatthrough.add(Blocks.double_plant);
	    		floatthrough.add(Blocks.wheat);
	    		floatthrough.add(Blocks.pumpkin);
	    		floatthrough.add(Blocks.melon_block);
	    		floatthrough.add(Blocks.pumpkin_stem);
	    		floatthrough.add(Blocks.melon_stem);
	    		floatthrough.add(Blocks.waterlily);
	    		floatthrough.add(Blocks.reeds);
	    		floatthrough.add(Blocks.cactus);
	    		floatthrough.add(Blocks.potatoes);
	    		floatthrough.add(Blocks.tallgrass);
			}
			if( Portfolio.config.get("blueprint", "floatThroughDecorations", "true").getBoolean(true) ) {
	       		//if( ! Portfolio.quiet ) { logger.info("Adding decorations to floatable block list"); }
	    		floatthrough.add(Blocks.rail); // treat decorations as air
	    		floatthrough.add(Blocks.torch);
	    		floatthrough.add(Blocks.redstone_torch);	    		
	    		floatthrough.add(Blocks.redstone_wire);
//	    		floatthrough.add(Blocks.wall_banner);
//	    		floatthrough.add(Blocks.standing_banner);
	    		floatthrough.add(Blocks.carpet);
	    		floatthrough.add(Blocks.skull);
	    		floatthrough.add(Blocks.tripwire);
	    		floatthrough.add(Blocks.lever);
	    		floatthrough.add(Blocks.stone_button);
	    		floatthrough.add(Blocks.wooden_button);
			}
			if( Portfolio.config.get("blueprint", "floatThroughFire", "true").getBoolean(true) ) {
	       		//if( ! Portfolio.quiet ) { logger.info("Adding fire to floatable block list"); }
				floatthrough.add(Blocks.fire);
			}
			if( Portfolio.config.get("blueprint", "floatThroughSnow", "true").getBoolean(true) ) {
	       		//if( ! Portfolio.quiet ) { logger.info("Adding snow to floatable block list"); }
				floatthrough.add(Blocks.snow);
				floatthrough.add(Blocks.snow_layer);
			}
			if( ! this.dry || Portfolio.config.get("blueprint", "floatThroughLiquids", "false").getBoolean(false) ) {  
				// treat liquids as air
	       		//if( ! Portfolio.quiet ) { logger.info("Adding liquids to floatable block list"); }
	    		floatthrough.add(Blocks.lava);
	    		floatthrough.add(Blocks.flowing_lava);
	    		floatthrough.add(Blocks.water);
	    		floatthrough.add(Blocks.flowing_water);    			
			}
			if( Portfolio.config.get("blueprint", "floatThroughIce", "false").getBoolean(false) ) {
				floatthrough.add(Blocks.ice);
			}
		}
    	
    	if( (this.flatWidth > 0) && (this.flatLength > 0) ) {
    		List<Integer> h = new LinkedList<Integer>();
    		int flattenTo;
    		IBlockState blockstate;
    		IBlockState newstate;
    		Block curBlock;
    		int x,y,z,w,l;
    		if( ! Portfolio.quiet ) {
    			logger.info("Flattenning area of plan: " + this.flatWidth + "x" + this.flatLength );
    		}
    		// Do the flattening, over location.getX+this.flatX --> location.getX+this.flatX+this.flatWidth, etc
    		// get the heights of all blocks in the flatzone.
    		for( w = 0; w < this.flatWidth; w++ ) {
        		for( l = 0; l < this.flatLength; l++ ) {
        			BlockPos bp;
        			x = this.flatX + w;
        			z = this.flatZ + l;
        			for( y = location.getY(); y < 256; y++ ) {  // up until in the air
            			bp = Directive.rotateAndAdd(new BlockPos(x,y,z),r,location);
            			bp = new BlockPos(bp.getX(),y,bp.getZ());
        	    		blockstate = Util.getBlockState(world,bp);
        	    		curBlock = blockstate.getBlock();
        				if( floatthrough.contains(curBlock) ) { break; }
        			}
        			for( ;y > 0; y-- ) {  // down until something solid beneath
            			bp = Directive.rotateAndAdd(new BlockPos(x,y-1,z),r,location);
            			bp = new BlockPos(bp.getX(),y-1,bp.getZ());
        	    		blockstate = Util.getBlockState(world,bp);
        	    		curBlock = blockstate.getBlock();
        				if( ! floatthrough.contains(curBlock) ) { break; }
        			}
        			h.add(y);
        		}
    			
    		}
    		// take the median
    		Collections.sort(h);
    		int pos = (int) Math.floor(h.size() / 2);
    		flattenTo = h.get(pos);
    		if( ! Portfolio.quiet ) { 
    			logger.info("Elements: " + h.size() + " Median: "+ pos );
    			logger.info("Low: " + h.get(0) + " High: "+ h.get(h.size()-1));     		
    		    logger.info("Median flattening height is: " + flattenTo); 
    		    logger.info("Maximum flattening is: " + this.flatMax); 
    		}
    		// bring down or build up every block in the zone
			if( this.flatMax < 1 ) { this.flatMax = 255; }
    		for( w = 0; w < this.flatWidth; w++ ) {
        		for( l = 0; l < this.flatLength; l++ ) {
        			BlockPos bp;
        			x = this.flatX + w;
        			z = this.flatZ + l;
        			newstate = Util.getState(Blocks.air, 0);
        			for( y = flattenTo; (y < 256)&&(y<(flattenTo+this.flatMax)); y++ ) {
            			bp = Directive.rotateAndAdd(new BlockPos(x,y,z),r,location);
            			bp = new BlockPos(bp.getX(),y,bp.getZ());
        	    		blockstate = Util.getBlockState(world,bp);
        	    		curBlock = blockstate.getBlock();        				
        				if( curBlock == Blocks.air ) {
        					break;
        				}
        				Util.setBlockState(world,bp,newstate);
        			}

        			if( flattenTo > 1 ) {
        				curBlock = Blocks.air;
        				blockstate = Util.getState(curBlock,0);
	        			for( y = flattenTo-1; (y > 0)&&(y > (flattenTo-this.flatMax)) ; y-- ) {
	            			bp = Directive.rotateAndAdd(new BlockPos(x,y-1,z),r,location);
	            			bp = new BlockPos(bp.getX(),y-1,bp.getZ());
	        	    		blockstate = Util.getBlockState(world,bp);
	        	    		curBlock = blockstate.getBlock();
	        				if( ! floatthrough.contains(curBlock) ) { break; }
	        			}
	        			if(y<1) { 
	        				blockstate = Util.getState(Blocks.stone, 0); 
        				} else if ( curBlock == Blocks.grass ) { // don't build up with grass on top of grass
	            			blockstate = Util.getState(Blocks.dirt,0);
	        			}
	        			for(  ; y < flattenTo; y++ ) {
	            			bp = Directive.rotateAndAdd(new BlockPos(x,y,z),r,location);
	            			bp = new BlockPos(bp.getX(),y,bp.getZ());
	        				Util.setBlockState(world,bp,blockstate);         				
	        			}
        			}
        		}
    		}
    	}
    	
    	// floating option -- go up.down to find surface of world
    	if( this.floaty ) {
    		BlockPos blk = new BlockPos(location);
    		
    		// float down to ground
    		Block b = Util.getBlock(world,location.down());
       		if( ! Portfolio.quiet ) { logger.info("Starting at: "+location.getY()+": under is "+Block.blockRegistry.getNameForObject(b)); }
    		while( location.getY() > 0 && floatthrough.contains(b) ) {
           		if( ! Portfolio.quiet ) { logger.info("Moving down from: "+location.getY()+" into "+Block.blockRegistry.getNameForObject(b)); }
    			location = location.down();
    			b = Util.getBlock(world,location.down());
    		}
       		if( ! Portfolio.quiet ) { logger.info("Stopped going down at: "+location.getY()+": under is "+Block.blockRegistry.getNameForObject(b)); }
    		// float up to air
			b = Util.getBlock(world,location);
       		while( location.getY() < 256 && ! floatthrough.contains(b) ) {
    			location = location.up(); 
           		if( ! Portfolio.quiet ) { logger.info("Moving up to: "+location.getY()+" from "+Block.blockRegistry.getNameForObject(b)); }
       			b = Util.getBlock(world,location);
    		}
       		if( ! Portfolio.quiet ) { logger.info("Stopped going up at: "+location.getY()+" in "+Block.blockRegistry.getNameForObject(b)); }
       		if( ! Portfolio.quiet ) { logger.info("Floating Y coordinate to surface level: "+location.getY()); }
    	}
    	
		Iterator i = this.steps.iterator();
	    while(i.hasNext()) {
	        Directive d = (Directive)i.next();
	    	//CheshireCat.logger.debug("Calling step: "+d.asText());
	        rv = d.build(world, location, r, this.logger,this.softlist);
	        if(!rv) {
	        	logger.error("Build of '"+this.name+"' failed with step: ["+d.asText()+"]");
	        	break; 
	        }
	    }
    	//CheshireCat.logger.info("Build of blueprint '"+this.name+"' ended");
		return rv;
	}
	public String asText() {
		String s = "";
		Iterator i = this.steps.iterator();
		if( visible ) {
			s = "PUBLIC";
		} else {
			s = "PRIVATE";
		}
		s = s + " PLAN " + name + " " + helptext + "\n";

	    while(i.hasNext()) {
	        Directive d = (Directive)i.next();
	        s = s + "  " + d.asText() + "\n";
	    }
	    s = s + "END\n";
		return s;
	}
	public boolean save(String f) {
		boolean rv = true;
		if( f == null ) { return false; }
		Writer writer = null;
		try {
		    writer = new BufferedWriter(new OutputStreamWriter(
		          new FileOutputStream(f), "utf-8")
		    );
		    writer.write("# Blueprint for: "+this.name+"\n");
		    writer.write("# "+this.helptext+"\n");
		    writer.write("# "+this.steps.size()+" steps.\n");
			writer.write(this.asText()+"\n");
		    writer.write("# End of blueprint\n");
		} catch (IOException ex) {
			logger.error("Unable to write file "+f+": "+ex.getMessage());
			rv = false;
		} finally {
		    try {writer.close();} catch (Exception ex) {}
		}
		return rv;
	}
}
